# Note: Tests for the TitlesAlignments class are in blast/test_titles.py
#       and diamond/test_titles.py because that class needs a concrete
#       (iterable) dark.alignments.ReadsAlignments class passed to its
#       __init__.  The tests below test the simpler dark.titles classes,
#       TitleAlignment and TitleAlignments.

from collections import Counter
import six
import warnings
import platform
from unittest import TestCase

from dark.titles import TitleAlignment, TitleAlignments
from dark.reads import Read
from dark.hsp import HSP, LSP

_pypy = platform.python_implementation() == 'PyPy'


class WarningTestMixin(object):
    """
    Provide an assertion test which checks to see that a specified warning
    was raised.
    """
    # Taken from
    # http://stackoverflow.com/questions/3892218/
    # how-to-test-with-pythons-unittest-that-a-warning-has-been-thrown

    def assertWarns(self, warning, callable, *args, **kwds):
        with warnings.catch_warnings(record=True) as warning_list:
            warnings.simplefilter('always')
            callable(*args, **kwds)
            self.assertTrue(
                any(item.category == warning for item in warning_list))


class TestTitleAlignment(TestCase):
    """
    Test the TitleAlignment class.
    """

    def testExpectedAttributes(self):
        """
        An instance of TitleAlignment must have the expected attributes.
        """
        read = Read('id', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        self.assertEqual(read, titleAlignment.read)
        self.assertEqual([], titleAlignment.hsps)

    def testToDict(self):
        """
        The toDict method must return the expected result.
        """
        read = Read('the-id', 'AAA')
        hsp1 = HSP(0, readStart=1, readEnd=2,
                   readStartInSubject=3, readEndInSubject=4,
                   subjectStart=5, subjectEnd=6,
                   readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                   readFrame=7, subjectFrame=8, identicalCount=9,
                   percentIdentical=31.3, positiveCount=10,
                   percentPositive=4.5)
        hsp2 = HSP(10, readStart=11, readEnd=12,
                   readStartInSubject=13, readEndInSubject=14,
                   subjectStart=15, subjectEnd=16,
                   readMatchedSequence='ggg', subjectMatchedSequence='ttt',
                   readFrame=17, subjectFrame=18, identicalCount=19,
                   percentIdentical=32.3, positiveCount=20,
                   percentPositive=4.6)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])

        self.assertEqual(
            {
                'hsps': [
                    {
                        'score': 0,
                        'readStart': 1,
                        'readEnd': 2,
                        'readStartInSubject': 3,
                        'readEndInSubject': 4,
                        'subjectStart': 5,
                        'subjectEnd': 6,
                        'readFrame': 7,
                        'subjectFrame': 8,
                        'identicalCount': 9,
                        'percentIdentical': 31.3,
                        'positiveCount': 10,
                        'percentPositive': 4.5,
                        'readMatchedSequence': 'aaa',
                        'subjectMatchedSequence': 'ccc',
                    },
                    {
                        'score': 10,
                        'readStart': 11,
                        'readEnd': 12,
                        'readStartInSubject': 13,
                        'readEndInSubject': 14,
                        'subjectStart': 15,
                        'subjectEnd': 16,
                        'readFrame': 17,
                        'subjectFrame': 18,
                        'identicalCount': 19,
                        'percentIdentical': 32.3,
                        'positiveCount': 20,
                        'percentPositive': 4.6,
                        'readMatchedSequence': 'ggg',
                        'subjectMatchedSequence': 'ttt',
                    },
                ],
                'read': {
                    'id': 'the-id',
                    'quality': None,
                    'sequence': 'AAA',
                },
            },
            titleAlignment.toDict())


class TestTitleAlignments(WarningTestMixin, TestCase):
    """
    Test the TitleAlignments class.
    """

    def testExpectedAttributes(self):
        """
        An instance of TitleAlignments must have the expected attributes.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        self.assertEqual('subject title', titleAlignments.subjectTitle)
        self.assertEqual(55, titleAlignments.subjectLength)
        self.assertEqual([], titleAlignments)

    def testAddAlignment(self):
        """
        It must be possible to add an alignment to an instance of
        TitleAlignments.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(read, titleAlignments[0].read)
        self.assertEqual([], titleAlignments[0].hsps)

    def testHSPs(self):
        """
        The hsps function must produce a list of all HSPs.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual([7, 14, 21],
                         [hsp.score.score for hsp in titleAlignments.hsps()])

    def testReadsEmpty(self):
        """
        The reads function must return an empty Reads instance if there are no
        reads for the title.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        self.assertEqual(0, len(list(titleAlignments.reads())))

    def testReads(self):
        """
        The reads function must return a Reads instance with the reads for
        the title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read1 = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read1, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read2 = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read2, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual([read1, read2], list(titleAlignments.reads()))

    def testReadCountZero(self):
        """
        The readCount function must return zero if no reads matched a title.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        self.assertEqual(0, titleAlignments.readCount())

    def testReadCount(self):
        """
        The readCount function must indicate how many reads matched a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(2, titleAlignments.readCount())

    def testHspCountZero(self):
        """
        The hspCount function must return zero if no reads matched a title.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        self.assertEqual(0, titleAlignments.hspCount())

    def testHspCount(self):
        """
        The hspCount function must indicate how many HSPs were found in
        total for all the alignments to a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(14)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(3, titleAlignments.hspCount())

    def testMedianScoreWithNoAlignments(self):
        """
        The medianScore function must raise IndexError (due to no inputs)
        if there are no alignments matching a title.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        error = '^arg is an empty sequence$'
        six.assertRaisesRegex(self, ValueError, error,
                              titleAlignments.medianScore)

    def testMedianScoreWithNoHsps(self):
        """
        The medianScore function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        error = '^arg is an empty sequence$'
        six.assertRaisesRegex(self, ValueError, error,
                              titleAlignments.medianScore)

    def testMedianScoreOfTwo(self):
        """
        The medianScore function must return the median score for the HSPs in
        all the alignments matching a title when given 2 scores.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(11, titleAlignments.medianScore())

    def testMedianScoreOfThree(self):
        """
        The medianScore function must return the median score for the HSPs in
        all the alignments matching a title when given 3 scores.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(15, titleAlignments.medianScore())

    def testBestHspWithNoHsps(self):
        """
        The bestHsp function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        if _pypy:
            error = '^arg is an empty sequence$'
        else:
            error = '^max\\(\\) arg is an empty sequence$'
        six.assertRaisesRegex(self, ValueError, error, titleAlignments.bestHsp)

    def testBestHsp(self):
        """
        The bestHsp function must return the HSP with the best score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp3, titleAlignments.bestHsp())

    def testWorstHspWithNoHsps(self):
        """
        The worstHsp function must raise ValueError if there are no HSPs.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [])
        titleAlignments.addAlignment(titleAlignment)
        if _pypy:
            error = '^arg is an empty sequence$'
        else:
            error = '^min\\(\\) arg is an empty sequence$'
        six.assertRaisesRegex(self, ValueError, error,
                              titleAlignments.worstHsp)

    def testWorstHsp(self):
        """
        The worstHsp function must return the HSP with the worst score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp1, titleAlignments.worstHsp())

    def testBetterThanFalse(self):
        """
        The hasScoreBetterThan function must return False if there is no HSP
        with a score better than the passed value.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertFalse(titleAlignments.hasScoreBetterThan(21))

    def testBetterThanTrue(self):
        """
        The hasScoreBetterThan function must return True if there is an HSP
        with a score better than the passed value.
        """
        hsp1 = HSP(7)
        hsp2 = HSP(15)
        hsp3 = HSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertTrue(titleAlignments.hasScoreBetterThan(20))

    def testCoverageNoReads(self):
        """
        The coverage method must return zero when a title alignments has no
        alignments (and therefore no coverage).
        """
        titleAlignments = TitleAlignments('subject title', 100)
        self.assertEqual(0.0, titleAlignments.coverage())

    def testFullCoverage(self):
        """
        The coverage method must return the correct value when the title is
        fully covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=0, subjectEnd=50)
        hsp2 = HSP(8, subjectStart=50, subjectEnd=100)
        titleAlignments = TitleAlignments('subject title', 100)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(1.0, titleAlignments.coverage())

    def testPartialCoverage(self):
        """
        The coverage method must return the correct value when the title is
        partially covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=10, subjectEnd=20)
        hsp2 = HSP(15, subjectStart=30, subjectEnd=40)
        hsp3 = HSP(21, subjectStart=50, subjectEnd=60)
        titleAlignments = TitleAlignments('subject title', 100)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(0.3, titleAlignments.coverage())

    def testFullCoverageCounts(self):
        """
        The coverageCounts method must return the correct result when the title
        is fully covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=0, subjectEnd=5)
        hsp2 = HSP(8, subjectStart=5, subjectEnd=10)
        titleAlignments = TitleAlignments('subject title', 10)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageCounts(self):
        """
        The coverageCounts method must return the correct results when the
        title is partially covered by its reads.
        """
        hsp1 = HSP(7, subjectStart=1, subjectEnd=2)
        hsp2 = HSP(15, subjectStart=3, subjectEnd=4)
        hsp3 = HSP(21, subjectStart=5, subjectEnd=6)
        titleAlignments = TitleAlignments('subject title', 10)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([1, 3, 5])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageCountsOverlap(self):
        """
        The coverageCounts method must return the correct results when the
        title is partially covered by its reads that overlap.
        """
        hsp1 = HSP(7, subjectStart=1, subjectEnd=2)
        hsp2 = HSP(15, subjectStart=3, subjectEnd=6)
        hsp3 = HSP(21, subjectStart=5, subjectEnd=6)
        titleAlignments = TitleAlignments('subject title', 10)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        c = Counter([1, 3, 4, 5, 5])
        self.assertEqual(c, titleAlignments.coverageCounts())

    def testCoverageInfoNoReads(self):
        """
        When a title has no reads aligned to it, the coverageInfo method
        must return an empty result.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual({}, coverage)

    def testCoverageInfoOneReadWithOneHSP(self):
        """
        When a title has one read with one HSP aligned to it, the coverageInfo
        method must return just the indices and bases from that read.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        hsp = HSP(15, subjectStart=3, subjectEnd=6, readMatchedSequence='CGT')
        read = Read('id1', 'AAACGT')
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                3: [(15, 'C')],
                4: [(15, 'G')],
                5: [(15, 'T')],
            },
            coverage)

    def testCoverageInfoOneReadWithTwoHSPs(self):
        """
        When a title has one read with two HSPs aligned to it, the coverageInfo
        method must return the correct indices and bases from that read.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        hsp1 = HSP(15, subjectStart=1, subjectEnd=4, readMatchedSequence='A-A')
        hsp2 = HSP(10, subjectStart=3, subjectEnd=6, readMatchedSequence='CGT')
        read = Read('id1', 'AAACGT')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                1: [(15, 'A')],
                2: [(15, '-')],
                3: [(15, 'A'), (10, 'C')],
                4: [(10, 'G')],
                5: [(10, 'T')],
            },
            coverage)

    def testCoverageInfoTwoReadsWithThreeHSPs(self):
        """
        When a title has two reads (one with two HSPs, one with one) aligned
        to it, the coverageInfo method must return the correct indices and
        bases from the read.
        """
        titleAlignments = TitleAlignments('subject title', 55)

        # First read.
        hsp1 = HSP(15, subjectStart=1, subjectEnd=4, readMatchedSequence='A-A')
        hsp2 = HSP(10, subjectStart=3, subjectEnd=6, readMatchedSequence='CGT')
        read = Read('id1', 'AAACGT')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)

        # Second read.
        hsp1 = HSP(20, subjectStart=5, subjectEnd=10,
                   readMatchedSequence='CGGTA')
        read = Read('id2', 'AAACGTCGGTAAAA')
        titleAlignment = TitleAlignment(read, [hsp1])
        titleAlignments.addAlignment(titleAlignment)

        coverage = titleAlignments.coverageInfo()
        self.assertEqual(
            {
                1: [(15, 'A')],
                2: [(15, '-')],
                3: [(15, 'A'), (10, 'C')],
                4: [(10, 'G')],
                5: [(10, 'T'), (20, 'C')],
                6: [(20, 'G')],
                7: [(20, 'G')],
                8: [(20, 'T')],
                9: [(20, 'A')],
            },
            coverage)

    def testResidueCountsNoReads(self):
        """
        When a title has no reads aligned to it, the residueCounts method
        must return an empty result.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        counts = titleAlignments.residueCounts()
        self.assertEqual(0, len(counts))

    def testResidueCountsUnknownCaseConversion(self):
        """
        The residueCounts method must raise a ValueError when asked to do an
        unknown case conversion.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        error = "convertCaseTo must be one of 'none', 'lower', or 'upper'"
        six.assertRaisesRegex(
            self, ValueError, error, titleAlignments.residueCounts,
            convertCaseTo='xxx')

    def testResidueCountsOneReadOneHSP(self):
        """
        The residueCounts method must return the correct result when just one
        read with one HSP is aligned to a title.
        """
        read = Read('id', 'ACGT')
        hsp = HSP(33, readStart=0, readEnd=4, readStartInSubject=0,
                  readEndInSubject=4, subjectStart=0, subjectEnd=4,
                  readMatchedSequence='ACGT', subjectMatchedSequence='ACGT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {'A': 1},
                1: {'C': 1},
                2: {'G': 1},
                3: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsOneReadOneHSPPartialMatch(self):
        """
        The residueCounts method must return the correct result when just one
        read with one HSP is aligned to a title and only part of the read
        matched the subject (all the read bases are still counted and
        returned).
        """
        read = Read('id', 'ACGT')
        hsp = HSP(33, readStart=0, readEnd=2, readStartInSubject=0,
                  readEndInSubject=4, subjectStart=0, subjectEnd=4,
                  readMatchedSequence='ACGT', subjectMatchedSequence='ACGT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {'A': 1},
                1: {'C': 1},
                2: {'G': 1},
                3: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsOneReadTwoHSPsAtStartOfSubject(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the leftmost HSP is
        aligned with the left edge of the subject.

        HSP1:       ACGT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=0,
                   readEndInSubject=4, subjectStart=0, subjectEnd=4,
                   readMatchedSequence='ACGT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=1,
                   readEndInSubject=5, subjectStart=1, subjectEnd=5,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {'A': 1},
                1: {'C': 2},
                2: {'G': 2},
                3: {'T': 2},
                4: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsOneReadTwoHSPsNotAtStartOfSubject(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the leftmost HSP is not
        aligned with the left edge of the subject.

        HSP1:       ACGT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='ACGT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=11,
                   readEndInSubject=15, subjectStart=11, subjectEnd=15,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {'A': 1},
                11: {'C': 2},
                12: {'G': 2},
                13: {'T': 2},
                14: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsNoCaseConversion(self):
        """
        The residueCounts method must return the correct result when asked not
        to convert case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='AcgT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=11,
                   readEndInSubject=15, subjectStart=11, subjectEnd=15,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {'A': 1},
                11: {'C': 1, 'c': 1},
                12: {'G': 1, 'g': 1},
                13: {'T': 2},
                14: {'T': 1},
            },
            titleAlignments.residueCounts(convertCaseTo='none'))

    def testResidueCountsCaseConvertLower(self):
        """
        The residueCounts method must return the correct result when asked to
        convert residues to lower case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='AcgT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=11,
                   readEndInSubject=15, subjectStart=11, subjectEnd=15,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {'a': 1},
                11: {'c': 2},
                12: {'g': 2},
                13: {'t': 2},
                14: {'t': 1},
            },
            titleAlignments.residueCounts(convertCaseTo='lower'))

    def testResidueCountsCaseConvertUpper(self):
        """
        The residueCounts method must return the correct result when asked to
        convert residues to upper case.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='AcgT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=11,
                   readEndInSubject=15, subjectStart=11, subjectEnd=15,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {'A': 1},
                11: {'C': 2},
                12: {'G': 2},
                13: {'T': 2},
                14: {'T': 1},
            },
            titleAlignments.residueCounts(convertCaseTo='upper'))

    def testResidueCountsCaseConvertUpperIsDefault(self):
        """
        The residueCounts method must convert to uppercase by default.

        HSP1:       AcgT
        HSP2:        CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='AcgT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=11,
                   readEndInSubject=15, subjectStart=11, subjectEnd=15,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                10: {'A': 1},
                11: {'C': 2},
                12: {'G': 2},
                13: {'T': 2},
                14: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsTwoReadsTwoHSPsLeftOverhang(self):
        """
        The residueCounts method must return the correct result when two
        reads, each with one HSP are aligned to a title and the leftmost HSP
        is aligned before the left edge of the subject (i.e, will include
        negative subject offsets).

        Subject:      GTT
        HSP1:       ACGT
        HSP2:        CGTT
        """
        read1 = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=-2,
                   readEndInSubject=2, subjectStart=0, subjectEnd=2,
                   readMatchedSequence='GT', subjectMatchedSequence='GT')
        read2 = Read('id', 'CGTT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=-1,
                   readEndInSubject=3, subjectStart=0, subjectEnd=3,
                   readMatchedSequence='GTT', subjectMatchedSequence='GTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read1, [hsp1])
        titleAlignments.addAlignment(titleAlignment)
        titleAlignment = TitleAlignment(read2, [hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                -2: {'A': 1},
                -1: {'C': 2},
                0: {'G': 2},
                1: {'T': 2},
                2: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testResidueCountsOneReadTwoHSPsNotOverlapping(self):
        """
        The residueCounts method must return the correct result when just one
        read with two HSPs is aligned to a title and the HSPs do not overlap
        one another.

        HSP1:    ACGT
        HSP2:              CGTT
        """
        read = Read('id', 'ACGT')
        hsp1 = HSP(33, readStart=0, readEnd=4, readStartInSubject=0,
                   readEndInSubject=4, subjectStart=0, subjectEnd=4,
                   readMatchedSequence='ACGT', subjectMatchedSequence='ACGT')
        hsp2 = HSP(33, readStart=0, readEnd=4, readStartInSubject=10,
                   readEndInSubject=14, subjectStart=10, subjectEnd=14,
                   readMatchedSequence='CGTT', subjectMatchedSequence='CGTT')
        titleAlignments = TitleAlignments('subject title', 55)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(
            {
                0: {'A': 1},
                1: {'C': 1},
                2: {'G': 1},
                3: {'T': 1},
                10: {'C': 1},
                11: {'G': 1},
                12: {'T': 1},
                13: {'T': 1},
            },
            titleAlignments.residueCounts())

    def testSummaryWhenEmpty(self):
        """
        If summary is called on an instance of TitleAlignments with no
        alignments a ValueError must be raised.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        error = '^max\\(\\) arg is an empty sequence$'
        six.assertRaisesRegex(self, ValueError, error, titleAlignments.summary)

    def testSummary(self):
        """
        The summary method must return the correct result.
        """
        titleAlignments = TitleAlignments('subject title', 10)
        titleAlignments.addAlignment(
            TitleAlignment(Read('id1', 'ACGT'), [
                HSP(30, subjectStart=0, subjectEnd=2),
            ]))
        titleAlignments.addAlignment(
            TitleAlignment(Read('id2', 'ACGT'), [
                HSP(55, subjectStart=2, subjectEnd=4),
                HSP(40, subjectStart=8, subjectEnd=9),
            ]))
        self.assertEqual(
            {
                'bestScore': 55,
                'coverage': 0.5,
                'hspCount': 3,
                'medianScore': 40,
                'readCount': 2,
                'subjectLength': 10,
                'subjectTitle': 'subject title',
            },
            titleAlignments.summary())

    def testToDict(self):
        """
        The toDict method must return the expected result.
        """
        read = Read('the-id', 'AAA')
        hsp1 = HSP(0, readStart=1, readEnd=2,
                   readStartInSubject=3, readEndInSubject=4,
                   subjectStart=5, subjectEnd=6,
                   readMatchedSequence='aaa', subjectMatchedSequence='ccc',
                   readFrame=7, subjectFrame=8, identicalCount=9,
                   percentIdentical=17.9, positiveCount=10,
                   percentPositive=3.9)
        hsp2 = HSP(10, readStart=11, readEnd=12,
                   readStartInSubject=13, readEndInSubject=14,
                   subjectStart=15, subjectEnd=16,
                   readMatchedSequence='ggg', subjectMatchedSequence='ttt',
                   readFrame=17, subjectFrame=18, identicalCount=19,
                   percentIdentical=27.9, positiveCount=20,
                   percentPositive=3.8)
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments = TitleAlignments('subject title', 10)
        titleAlignments.addAlignment(titleAlignment)

        self.assertEqual(
            {
                'subjectTitle': 'subject title',
                'subjectLength': 10,
                'titleAlignments': [
                    {
                        'hsps': [
                            {
                                'score': 0,
                                'readStart': 1,
                                'readEnd': 2,
                                'readStartInSubject': 3,
                                'readEndInSubject': 4,
                                'subjectStart': 5,
                                'subjectEnd': 6,
                                'readFrame': 7,
                                'subjectFrame': 8,
                                'identicalCount': 9,
                                'percentIdentical': 17.9,
                                'positiveCount': 10,
                                'percentPositive': 3.9,
                                'readMatchedSequence': 'aaa',
                                'subjectMatchedSequence': 'ccc',
                            },
                            {
                                'score': 10,
                                'readStart': 11,
                                'readEnd': 12,
                                'readStartInSubject': 13,
                                'readEndInSubject': 14,
                                'subjectStart': 15,
                                'subjectEnd': 16,
                                'readFrame': 17,
                                'subjectFrame': 18,
                                'identicalCount': 19,
                                'percentIdentical': 27.9,
                                'positiveCount': 20,
                                'percentPositive': 3.8,
                                'readMatchedSequence': 'ggg',
                                'subjectMatchedSequence': 'ttt',
                            },
                        ],
                        'read': {
                            'id': 'the-id',
                            'quality': None,
                            'sequence': 'AAA',
                        },
                    },
                ],
            },
            titleAlignments.toDict())


class TestTitleAlignmentsLSP(TestCase):
    """
    Test the TitleAlignments class using LSPs. The only tests here are ones
    that depend on lower scores being better.
    """

    def testBestHsp(self):
        """
        The bestHsp function must return the HSP with the best score for the
        HSPs all the alignments matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp1, titleAlignments.bestHsp())

    def testWorstHsp(self):
        """
        The worstHsp function must return the HSP with the worst score for all
        the HSPs for all the alignments matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(hsp3, titleAlignments.worstHsp())

    def testBetterThanFalse(self):
        """
        The hasScoreBetterThan function must return False if there is no HSP
        with a score better than the passed value.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertFalse(titleAlignments.hasScoreBetterThan(7))

    def testBetterThanTrue(self):
        """
        The hasScoreBetterThan function must return True if there is an HSP
        with a score better than the passed value.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertTrue(titleAlignments.hasScoreBetterThan(9))

    def testReadIdsEmpty(self):
        """
        The readIds function must return the empty set if no reads matched a
        title.
        """
        titleAlignments = TitleAlignments('subject title', 55)
        self.assertEqual(0, len(titleAlignments.readIds()))

    def testReadIds(self):
        """
        The readIds function must return the set of read ids for the alignments
        matching a title.
        """
        hsp1 = LSP(7)
        hsp2 = LSP(15)
        hsp3 = LSP(21)
        titleAlignments = TitleAlignments('subject title', 55)
        read = Read('id1', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp1, hsp2])
        titleAlignments.addAlignment(titleAlignment)
        read = Read('id2', 'AAA')
        titleAlignment = TitleAlignment(read, [hsp3])
        titleAlignments.addAlignment(titleAlignment)
        self.assertEqual(set(['id1', 'id2']), titleAlignments.readIds())



from __future__ import division

from collections import defaultdict

from dark.utils import countPrint

try:
    from itertools import zip_longest
except ImportError:
    # zip_longest does not exist in Python 2.7 itertools. We should be able
    # to get it via from six.moves import zip_longest according to
    # https://pythonhosted.org/six/index.html?highlight=zip_longest but
    # that doesn't work for me.
    from itertools import izip_longest as zip_longest

# A list of the ambiguous values is given at
# https://en.wikipedia.org/wiki/Nucleic_acid_notation
AMBIGUOUS = {
    'A': {'A'},
    'C': {'C'},
    'G': {'G'},
    'T': {'T'},
    'M': {'A', 'C'},
    'R': {'A', 'G'},
    'W': {'A', 'T'},
    'S': {'G', 'C'},
    'K': {'G', 'T'},
    'Y': {'C', 'T'},
    'V': {'A', 'C', 'G'},
    'H': {'A', 'C', 'T'},
    'D': {'A', 'G', 'T'},
    'B': {'C', 'G', 'T'},
    'N': {'A', 'C', 'G', 'T'},
}

# Make a reverse version of AMBIGUOUS.
BASES_TO_AMBIGUOUS = dict(
    (''.join(sorted(bases)), symbol) for symbol, bases in AMBIGUOUS.items())


def matchToString(dnaMatch, read1, read2, matchAmbiguous=True, indent='',
                  offsets=None, includeGapLocations=True):
    """
    Format a DNA match as a string.

    @param dnaMatch: A C{dict} returned by C{compareDNAReads}.
    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, ambiguous nucleotides that are
        possibly correct were counted as actually being correct. Otherwise,
        the match was done strictly, insisting that only non-ambiguous
        nucleotides could contribute to the matching nucleotide count.
    @param indent: A C{str} to indent all returned lines with.
    @param offsets: If not C{None}, a C{set} of offsets of interest that were
        only considered when making C{match}.
    @param includeGapLocations: If C{True} indicate the (1-based) locations of
        gaps.
    @return: A C{str} describing the match.
    """
    match = dnaMatch['match']
    identicalMatchCount = match['identicalMatchCount']
    ambiguousMatchCount = match['ambiguousMatchCount']
    gapMismatchCount = match['gapMismatchCount']
    gapGapMismatchCount = match['gapGapMismatchCount']
    nonGapMismatchCount = match['nonGapMismatchCount']

    if offsets:
        len1 = len2 = len(offsets)
    else:
        len1, len2 = map(len, (read1, read2))

    result = []
    append = result.append

    append(countPrint('%sExact matches' % indent, identicalMatchCount,
                      len1, len2))
    append(countPrint('%sAmbiguous matches' % indent, ambiguousMatchCount,
                      len1, len2))
    if ambiguousMatchCount and identicalMatchCount:
        anyMatchCount = identicalMatchCount + ambiguousMatchCount
        append(countPrint('%sExact or ambiguous matches' % indent,
                          anyMatchCount, len1, len2))

    mismatchCount = (gapMismatchCount + gapGapMismatchCount +
                     nonGapMismatchCount)
    append(countPrint('%sMismatches' % indent, mismatchCount, len1, len2))
    conflicts = 'conflicts' if matchAmbiguous else 'conflicts or ambiguities'
    append(countPrint('%s  Not involving gaps (i.e., %s)' % (indent,
                      conflicts), nonGapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in one sequence' % indent,
                      gapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in both sequences' % indent,
                      gapGapMismatchCount, len1, len2))

    for read, key in zip((read1, read2), ('read1', 'read2')):
        append('%s  Id: %s' % (indent, read.id))
        length = len(read)
        append('%s    Length: %d' % (indent, length))
        gapCount = len(dnaMatch[key]['gapOffsets'])
        append(countPrint('%s    Gaps' % indent, gapCount, length))
        if includeGapLocations and gapCount:
            append(
                '%s    Gap locations (1-based): %s' %
                (indent,
                 ', '.join(map(lambda offset: str(offset + 1),
                               sorted(dnaMatch[key]['gapOffsets'])))))
        ambiguousCount = len(dnaMatch[key]['ambiguousOffsets'])
        append(countPrint('%s    Ambiguous' % indent, ambiguousCount, length))
        extraCount = dnaMatch[key]['extraCount']
        if extraCount:
            append(countPrint('%s    Extra nucleotides at end' % indent,
                              extraCount, length))

    return '\n'.join(result)


def compareDNAReads(read1, read2, matchAmbiguous=True, gapChars='-',
                    offsets=None):
    """
    Compare two DNA sequences.

    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct, and score these in the
        ambiguousMatchCount. Otherwise, we are strict and insist that only
        non-ambiguous nucleotides can contribute to the matching nucleotide
        count.
    @param gapChars: An object supporting __contains__ with characters that
        should be considered to be gaps.
    @param offsets: If not C{None}, a C{set} of offsets of interest. Offsets
        not in the set will not be considered.
    @return: A C{dict} with information about the match and the individual
        sequences (see below).
    """
    identicalMatchCount = ambiguousMatchCount = 0
    gapMismatchCount = nonGapMismatchCount = gapGapMismatchCount = 0
    read1ExtraCount = read2ExtraCount = 0
    read1GapOffsets = []
    read2GapOffsets = []
    read1AmbiguousOffsets = []
    read2AmbiguousOffsets = []
    empty = set()

    def _identicalMatch(a, b):
        return a == b and len(AMBIGUOUS[a]) == 1

    def _ambiguousMatch(a, b, matchAmbiguous):
        """
        Checks if two characters match ambiguously if matchAmbiguous is True.
        A match is an ambiguous match if it is not an identical match, but the
        sets of ambiguous characters overlap.
        """
        return (matchAmbiguous and
                not _identicalMatch(a, b) and
                AMBIGUOUS.get(a, empty) & AMBIGUOUS.get(b, empty))

    for offset, (a, b) in enumerate(zip_longest(read1.sequence.upper(),
                                                read2.sequence.upper())):
        # Use 'is not None' in the following to allow an empty offsets set
        # to be passed.
        if offsets is not None and offset not in offsets:
            continue
        if len(AMBIGUOUS.get(a, '')) > 1:
            read1AmbiguousOffsets.append(offset)
        if len(AMBIGUOUS.get(b, '')) > 1:
            read2AmbiguousOffsets.append(offset)
        if a is None:
            # b has an extra character at its end (it cannot be None).
            assert b is not None
            read2ExtraCount += 1
            if b in gapChars:
                read2GapOffsets.append(offset)
        elif b is None:
            # a has an extra character at its end.
            read1ExtraCount += 1
            if a in gapChars:
                read1GapOffsets.append(offset)
        else:
            # We have a character from both sequences (they could still be
            # gap characters).
            if a in gapChars:
                read1GapOffsets.append(offset)
                if b in gapChars:
                    # Both are gaps. This can happen (though hopefully not
                    # if the sequences were pairwise aligned).
                    gapGapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # a is a gap, b is not.
                    gapMismatchCount += 1
            else:
                if b in gapChars:
                    # b is a gap, a is not.
                    gapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # Neither is a gap character.
                    if _identicalMatch(a, b):
                        identicalMatchCount += 1
                    elif _ambiguousMatch(a, b, matchAmbiguous):
                        ambiguousMatchCount += 1
                    else:
                        nonGapMismatchCount += 1

    return {
        'match': {
            'identicalMatchCount': identicalMatchCount,
            'ambiguousMatchCount': ambiguousMatchCount,
            'gapMismatchCount': gapMismatchCount,
            'gapGapMismatchCount': gapGapMismatchCount,
            'nonGapMismatchCount': nonGapMismatchCount,
        },
        'read1': {
            'ambiguousOffsets': read1AmbiguousOffsets,
            'extraCount': read1ExtraCount,
            'gapOffsets': read1GapOffsets,
        },
        'read2': {
            'ambiguousOffsets': read2AmbiguousOffsets,
            'extraCount': read2ExtraCount,
            'gapOffsets': read2GapOffsets,
        },
    }


def findKozakConsensus(read):
    """
    In a given DNA sequence, search for a Kozak consensus: (gcc)gccRccATGG.
    The upper case bases in that pattern are required, and the lower case
    bases are the ones most frequently found at the given positions. The
    initial 'gcc' sequence (in parentheses) is of uncertain significance
    and is not taken into account here.

    @param read: A C{DNARead} instance to be checked for Kozak consensi.
    @return: A generator that yields C{DNAKozakRead} instances.
    """
    from dark.reads import DNAKozakRead

    readLen = len(read)
    if readLen > 9:
        offset = 6
        readSeq = read.sequence
        while offset < readLen - 3:
            triplet = readSeq[offset:offset + 3]
            if triplet == 'ATG':
                if readSeq[offset + 3] == 'G':
                    if readSeq[offset - 3] in 'GA':
                        kozakQualityCount = sum((
                            readSeq[offset - 1] == 'C',
                            readSeq[offset - 2] == 'C',
                            readSeq[offset - 4] == 'C',
                            readSeq[offset - 5] == 'C',
                            readSeq[offset - 6] == 'G'))

                        kozakQualityPercent = kozakQualityCount / 5.0 * 100
                        yield DNAKozakRead(read, offset - 6, offset + 4,
                                           kozakQualityPercent)
            offset += 1


class FloatBaseCounts(object):
    """
    Hold a floating point count of possible nucleotide bases.

    @param codes: An iterable of nucleotide codes.
    @param unknownAreAmbiguous: If C{True}, any unknown character (e.g., a '-'
        gap or '?' unknown base) will be treated as being fully ambiguous
        (i.e., could be any of ACGT). Otherwise, all unknown characters are
        collected under the count for '-'.
    """
    def __init__(self, codes, unknownAreAmbiguous=False):
        self.codes = list(map(str.upper, codes))
        self.unknownAreAmbiguous = unknownAreAmbiguous
        self.n = len(self.codes)
        counts = defaultdict(float)

        default = self._default = set('ACGT') if unknownAreAmbiguous else {'-'}

        for code in self.codes:
            possible = AMBIGUOUS.get(code, default)
            frac = 1.0 / len(possible)
            for p in possible:
                counts[p] += frac

        # Sort first by base.
        _sorted = [(base, counts[base]) for base in sorted(counts)]

        # Then by count (reversed).
        def key(item):
            return item[1]

        self._sorted = sorted(_sorted, key=key, reverse=True)
        self.counts = counts

    def mostFrequent(self):
        """
        Which bases are the most frequent?

        @return: A C{set} of the most frequent bases.
        """
        maxCount = self._sorted[0][1]
        return set(base for base, count in self._sorted if count == maxCount)

    def highestFrequency(self):
        """
        What is the frequency of the most frequent base?

        @return: The C{float} frequency of the most common base.
        """
        if len(self.counts) < 2:
            return 1.0
        else:
            return self._sorted[0][1] / float(self.n)

    def homogeneous(self, level):
        """
        Does the most frequent base occurs at least C{level} fraction of the
        time?

        @param level: A C{float} fraction.
        @return: C{True} if the most common base occurs at least C{level}
            fraction of the time. If there are no bases at all, this is
            considered homogeneous.
        """
        return self.highestFrequency() >= level

    def __len__(self):
        return len(self.counts)

    def __str__(self):
        fmt = '%d' if all(c == int(c) for b, c in self._sorted) else '%.2f'
        return '%s (%.3f)' % (
            ' '.join(('%s:' + fmt) % (b, c) for b, c in self._sorted),
            self.highestFrequency())

    def variable(self, confirm=True):
        """
        Are the nucleotides variable?

        @param confirm: If C{True}, confirm that there is variability by
            looking at the ambiguous nucleotides. Else just report C{True}
            if there is more than one code (which might not indicate actual
            variability, since two codes could be ambiguous and have a
            nucleotide in common).
        """
        codes = self.codes

        if confirm:
            unambiguous = set()
            ambiguousIntersection = None
            default = self._default
            for code in codes:
                possible = AMBIGUOUS.get(code, default)
                if len(possible) == 1:
                    unambiguous.add(code)
                else:
                    if ambiguousIntersection is None:
                        ambiguousIntersection = set(possible)
                    else:
                        ambiguousIntersection.intersection_update(possible)

            if len(unambiguous) == 0:
                # There were no unambiguous nucleotide codes.

                # Sanity check: there must have been some ambiguous sites.
                assert ambiguousIntersection is not None

                if len(ambiguousIntersection) == 0:
                    # The ambiguous sites had nothing in common, so
                    # variation must exist (it cannot be determined what
                    # the variation is, but we don't care about that).
                    return True
                else:
                    # All the ambiguous sites have at least one nucleotide
                    # in common, so we can't be sure there's any variation.
                    pass
            elif len(unambiguous) == 1:
                # All the unambiguous sites agree. Do any of the ambiguous
                # sites (if any) not allow the unambiguous nucleotide in
                # their possibilities?
                if ambiguousIntersection is None:
                    # There were no ambiguous sites, so there's no
                    # variation here.
                    pass
                else:
                    # If any of the ambiguous sites excludes the single
                    # unambiguous nucleotide, then variation must exist.
                    nt = unambiguous.pop()
                    for code in codes:
                        possible = AMBIGUOUS.get(code, default)
                        if nt not in possible:
                            return True
            elif len(unambiguous) > 1:
                return True
        else:
            if len(codes) > 1:
                return True

        return False


def sequenceToRegex(sequence, wildcards='-?'):
    """
    Convert a potentially ambiguous DNA sequence into a regular expression.
    '?' and '-' are translated into [ACGT].

    @param sequence: A C{str} DNA sequence, possibly with ambiguous codes.
        Case insensitive.
    @param wildcards: A C{set} (or C{str}) with characters that should be
        translated to '[ACGT]'. Note that this happens only if the standard
        ambiguous lookup fails (the order could be changed one day if we need
        to override, or we could allow the passing of an ambiguity mapping).
        Wildcards are case sensitive.
    @raise KeyError: If any character in C{sequence} is unknown.
    @return: A C{str} regular expression with [...] for the ambiguous codes in
        C{sequence}.
    """
    result = []
    append = result.append
    for base in sequence.upper():
        try:
            possible = ''.join(sorted(AMBIGUOUS[base]))
        except KeyError:
            if base in wildcards:
                possible = 'ACGT'
            else:
                raise

        append(('[%s]' % possible) if len(possible) > 1 else possible)

    return ''.join(result)



#!/usr/bin/env python

"""
Read a file of our JSON BLAST output (which has one JSON object per
line) from stdin and pretty print it to stdout.
"""

from __future__ import print_function

from json import dumps, loads
import sys

for line in sys.stdin:
    s = dumps(loads(line[:-1]), sort_keys=True, indent=2)
    print('\n'.join([field.rstrip() for field in s.splitlines()]))



import warnings


class GenomeRanges(object):
    """
    Split and manage a GenBank range string as converted by BioPython.

    @param rangeStr: A C{str} indicating the (0-based) genome nucleotide
        offsets covered by a protein. This can have the following example
        forms:

        [9462:10137](+)
        [11969:12575](-)
        join{[126386:126881](-), [125941:126232](+)}
        join{[153005:153821](+), [82030:82618](-), [80480:81305](-)}

        Note that this is the string as returned from BioPython SeqIO.parse
        when it parses a GenBank flat file, not the string that is in such a
        file. For that use the C{splitRange} function you can find in an
        earlier version of this file (i.e., look in its git history).

    @raise ValueError: If C{rangeStr} cannot be correctly parsed.
    @return: A C{list} whose elements are 3-C{tuple}s holding the C{int} start
        and stop offsets and a C{bool} that indicates whether the offsets
        corresponds to the forward strand (i.e., will be C{True} when
        there is a '(+)' in a range and C{False} when there is a '(-)' which
        indicates the reverse strand).

        Example arguments and their return values:

            '[9462:10137](+)' => ((9462, 10137, True),)

            '[11969:12575](-)' => ((11969, 12575, False),)

            'join{[126386:126881](-), [125941:126232](+)}' =>
                ((126386, 126881, False),
                 (125941, 126232, True))
    """
    def __init__(self, rangeStr):
        if rangeStr.startswith('join{') and rangeStr[-1] == '}':
            join = True
            inner = rangeStr[5:-1]
        else:
            join = False
            inner = rangeStr

        ranges = []
        subRanges = inner.split(', ')
        nRanges = len(subRanges)

        if nRanges == 1 and join:
            raise ValueError('Could not parse GenBank range string "%s". '
                             'join{} can only be used with multiple ranges.' %
                             rangeStr)
        elif nRanges > 1 and not join:
            raise ValueError('Could not parse GenBank range string "%s". '
                             'Multiple ranges must be wrapped in join{}.' %
                             rangeStr)

        for subRange in subRanges:
            if subRange.endswith('](+)'):
                forward = True
            elif subRange.endswith('](-)'):
                forward = False
            else:
                raise ValueError('Could not parse GenBank range string "%s". '
                                 'Range "%s" does not end with ](+) or ](-).' %
                                 (rangeStr, subRange))

            if not subRange.startswith('['):
                raise ValueError('Could not parse GenBank range string "%s". '
                                 'Range "%s" does not start with "[".' %
                                 (rangeStr, subRange))
            try:
                start, stop = subRange[1:-4].split(':')
            except ValueError as e:
                raise ValueError('Could not parse GenBank range string "%s". '
                                 'Original parsing ValueError was "%s".' %
                                 (rangeStr, e))
            else:
                if start.startswith('<'):
                    start = start[1:]
                if stop.startswith('>'):
                    stop = stop[1:]
                start, stop = map(int, (start, stop))
                if start > stop:
                    raise ValueError(
                        'Could not parse GenBank range string "%s". '
                        'Offset values (%d, %d) cannot decrease.' %
                        (rangeStr, start, stop))

                ranges.append((start, stop, forward))

        self.ranges = self._mergeContiguousRanges(ranges)
        self._nRanges = len(self.ranges)

    def _mergeContiguousRanges(self, ranges):
        """
        Merge ranges that are contiguous (follow each other immediately on the
        genome).

        @param ranges: An iterable of (start, stop, forward) tuples.
        @return: A C{tuple} of (start, stop, forward) tuples with contiguous
            ranges found in C{ranges} merged.
        """
        result = []
        lastStart = lastStop = lastForward = None

        for index, (start, stop, forward) in enumerate(ranges):
            if lastStart is None:
                lastStart, lastStop, lastForward = start, stop, forward
            else:
                if start == lastStop and forward == lastForward:
                    # This range continues the previous one.
                    warnings.warn(
                        'Contiguous GenBank ranges detected: [%d:%d] '
                        'followed by [%d:%d].' %
                        (lastStart, lastStop, start, stop))
                    lastStop = stop
                else:
                    # Emit the range that just got terminated.
                    result.append((lastStart, lastStop, lastForward))
                    # And remember this one.
                    lastStart, lastStop, lastForward = start, stop, forward

        # Emit the final range.
        result.append((lastStart, lastStop, lastForward))

        return tuple(result)

    def __str__(self):
        return '<%s: %s>' % (
            self.__class__.__name__, ', '.join(map(str, self.ranges)))

    def circular(self, genomeLength):
        """
        Determine whether the offset ranges of a protein in a genome span the
        end of the genome (indicating that the genome may be circular).

        @param genomeLength: The C{int} length of the genome.
        @return: A C{bool}, C{True} if the ranges overlaps the genome end.
        """
        if self._nRanges == 1:
            # If there is only one range, we simply return False even though it
            # could be that the single range spans the whole genome. That is
            # not the same as having two ranges, one of which ends at the end
            # of the genome and the next starting at the start of the genome.
            #
            # I.e., a range tuples like this ((0, 100, False),) on a genome of
            # length 100 will return False, whereas ((0, 75, False), (75, 100,
            # False)) on a genome of length 100 will return True.
            #
            # The decision about whether you consider the degenerate case of
            # ((0, 100, False),) to indicate circularity is a bit
            # arbitrary. Here I've decided that it should not because I intend
            # to use this function to decide whether to output multiple SAM
            # lines for the match of a read against a reference. In this
            # degenerate case just a single SAM line suffices, whereas in a
            # non-degenerate case (such as ((0, 75, False), (75, 100, False))
            # on a genome of length 100), a chimeric SAM alignment must be
            # used, which requires quite different processing than a normal
            # linear match. YMMV, and if you don't like it you can easily add
            # an argument to this function that changes this behaviour (leaving
            # the default to do what the code currently does).
            return False

        for index, (start, stop, _) in enumerate(self.ranges):
            if stop == genomeLength:
                return self.ranges[(index + 1) % self._nRanges][0] == 0

        return False

    def startInGenome(self, match):
        """
        Calculate the position in the nucleotide genome where a DIAMOND match
        begins.

        @param match: A C{dict} with information about the DIAMOND match, as
            returned by C{self._preprocessMatch}. Must contain a 'sstart' key
            giving a 1-based C{int} offset of the start of the match in the
            protein.
        @return: The C{int} offset of the start of the DIAMOND match in the
            genome.
        """

        # Calculate the start of the match in the genome, given its start in
        # the protein.
        offsetInGenome = remaining = (match['sstart'] - 1) * 3

        for start, stop, _ in self.ranges:
            rangeWidth = stop - start
            if remaining < rangeWidth:
                # The match starts in this range.
                return start + remaining
            else:
                remaining -= rangeWidth
        else:
            raise ValueError(
                'Starting nucleotide offset %d not found in protein '
                'nucleotide ranges %s.' %
                (offsetInGenome,
                 ', '.join(('(%d, %d)' % (i, j))
                           for i, j, _ in self.ranges)))

    def orientations(self):
        """
        Produce the set of all orientations for our ranges.

        @return: A C{set} of C{True} and C{False} range orientations.
        """
        return set(r[2] for r in self.ranges)

    def distinctRangeCount(self, genomeLength):
        """
        Determine the number of distinct ranges, given the genome length.

        @param genomeLength: The C{int} length of the genome.
        @return: An C{int} which is the number of ranges that are not
            contiguous with one another.
        """
        return len(self.ranges) - int(self.circular(genomeLength))



#!/usr/bin/env python

from __future__ import print_function

import sys
import hashlib

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('Given FASTA on stdin, write it to stdout according to a '
                     'given format.'))

    parser.add_argument(
        '--prefix',
        help=('The string prefix to be used. E.g., use --prefix r for raw '
              'strings. The default is "f" (i.e., use Python f-strings) for '
              'Python 3.6 and later, or else "" (i.e., regular strings).'))

    parser.add_argument(
        '--start', default=1, type=int,
        help='The integer to start counting from, for the "count" variable.')

    parser.add_argument(
        '--end',
        help=('A string to use as the "end" keyword argument for print(). '
              'This will be passed to eval, allowing convenient use of '
              'Python string constructs such as \\t on the command line. '
              'If not specified, a newline will be printed.'))

    parser.add_argument(
        '--format',
        help=("The output format. The default is to reproduce the input. "
              "The format can be that used by Python to specify items in a "
              'dictionary. Available dictionary items are "id", "sequence", '
              '"quality", "length", "count", and "md5" (of the sequence). A '
              "dark.reads.Read instance (named 'read') and an 'md5' "
              "function (that takes and returns a string) are both also in "
              "scope for each FASTA record and these can be accessed using "
              "f-string format (with Python 3.6 or greater). The passed "
              "format string will be eval'd in a double quoted string, as "
              "with the --end string, allowing the use of Python string "
              "conveniences such as \\t on the command line."))

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    format_ = args.format
    end = '\n' if args.end is None else eval('"' + args.end + '"')

    if args.prefix is None:
        prefix = '' if sys.version_info < (3, 6) else 'f'
    else:
        prefix = args.prefix

    if format_ is None:
        format_ = (r'>%(id)s\n%(sequence)s' if args.fasta else
                   r'@%(id)s\n%(sequence)s\n+\n%(quality)s')

    needMd5 = '%(md5)' in format_
    needLength = '%(length)' in format_

    def md5(s):
        return hashlib.md5(s.encode('utf-8')).hexdigest()

    for count, read in enumerate(reads, start=args.start):
        sequence = read.sequence
        d = {
            'count': count,
            'id': read.id,
            'quality': read.quality,
            'sequence': sequence,
        }
        if needMd5:
            d['md5'] = md5(sequence)
        if needLength:
            d['length'] = len(sequence)

        print(eval(prefix + '"' + format_ + '" % d'), end=end)



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
from os.path import exists, basename

from dark.process import Executor

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Run bwa on a FASTA file. Optionally convert the result to '
                 'BAM, sorting, and indexing.'))

parser.add_argument(
    '--bwaIndex', required=True,
    help='The name of the BWA index (created with bwa index).')

parser.add_argument(
    '--bwaArgs', default='mem',
    help='All arguments to be passed to BWA.')

parser.add_argument(
    '--fastaFile', required=True,
    help=('The name of the FASTA file whose sequences should be aligned with '
          'the BWA index.'))

parser.add_argument(
    '--base',
    help=('The base name of files to create. Suffixes such as .sam and .bam '
          'will be added. If not given, the basename of the --fastaFile name '
          'will be used, stripped of its final suffix.'))

parser.add_argument(
    '--verbose', default=False, action='store_true',
    help='If specified, show the commands that were (or would be) executed.')

parser.add_argument(
    '--noBAM', default=False, action='store_true',
    help='If specified, do not convert SAM to BAM.')

parser.add_argument(
    '--noSort', default=False, action='store_true',
    help='If specified, do not sort the BAM.')

parser.add_argument(
    '--noIndex', default=False, action='store_true',
    help='If specified, do not index the sorted BAM.')

parser.add_argument(
    '--noClean', default=False, action='store_true',
    help='If specified, do not remove intermediate .sam and .bam files.')

parser.add_argument(
    '--force', default=False, action='store_true',
    help='If specified, overwrite pre-existing output files.')

parser.add_argument(
    '--dryRun', default=False, action='store_true',
    help='If specified, do not run commands, just print what would be done.')

args = parser.parse_args()

if args.base is None:
    fields = basename(args.fastaFile).rsplit('.', 1)
    if len(fields) < 2:
        print('No --base argument was given and the --fastaFile argument '
              'does not have a .suffix that can be stripped.',
              file=sys.stderr)
        sys.exit(1)
    base = fields[0]
else:
    base = args.base


samFile = base + '.sam'
bamFile = base + '.bam'
sortedBamFile = base + '-sorted.bam'

if not (args.force or args.dryRun):
    existing = []
    for filename in samFile, bamFile, sortedBamFile:
        if exists(filename):
            existing.append(filename)
    if existing:
        print('Will not overwrite pre-existing file%s %s. '
              'Use --force to make me.' % (
                  '' if len(existing) == 1 else 's',
                  ', '.join(existing)),
              file=sys.stderr)
        sys.exit(2)

e = Executor(args.dryRun)

e.execute("bwa %s '%s' '%s' > '%s'" % (
    args.bwaArgs, args.bwaIndex, args.fastaFile, samFile))

if not args.noBAM:
    e.execute("samtools view -b < '%s' > '%s'" % (samFile, bamFile))

    if not args.noClean:
        e.execute("rm '%s'" % samFile)

    if not args.noSort:
        e.execute("samtools sort '%s' > '%s'" % (bamFile, sortedBamFile))

        if not args.noClean:
            e.execute("rm '%s'" % bamFile)

        if not args.noIndex:
            e.execute("samtools index '%s'" % sortedBamFile)

if args.dryRun or args.verbose:
    print('\n'.join(e.log))



#!/usr/bin/env python

import os
import sys
import argparse
from tempfile import mkdtemp
from os.path import join

from dark.fasta import FastaReads
from dark.process import Executor


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Make a consensus sequence.')

    parser.add_argument(
        '--reference', required=True,
        help='The reference FASTA file.')

    parser.add_argument(
        '--bam',
        help=('The BAM file from which the consensus should be made. '
              'Required if --maskLowCoverage is used. If no BAM file is '
              'given, a VCF file must be provided. If both a BAM and a VCF '
              'file are given, the VCF file will take precedence.'))

    parser.add_argument(
        '--vcfFile',
        help=('The VCF file. If omitted, bcftools will be used to make a VCF '
              'file from the BAM file.'))

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        '--id',
        help=('The id to use in the consensus sequence in the output FASTA. '
              'If not given, the reference sequence id will be used.'))

    group.add_argument(
        '--idLambda', metavar='LAMBDA-FUNCTION',
        help=('A one-argument function taking and returning a read id. '
              'This can be used to set the id of the reference sequence based '
              'on the id of the reference sequence (the function will be '
              'called with the id of the reference sequence). E.g., '
              '--idLambda "lambda id: id.split(\'_\')[0]" or '
              '--idLambda "lambda id: id[:10] + \'-consensus\'".'))

    parser.add_argument(
        '--sample',
        help=('The name of the sample (from the @RG SM tag in the original '
              'alignment BAM file) for which a consensus should be made. '
              'If not given, the first sample name (from the #CHROM header) '
              'in the VCF file will be used.'))

    parser.add_argument(
        '--dryRun', default=False, action='store_true',
        help='Do not run commands, just print what would be done.')

    parser.add_argument(
        '--maskLowCoverage', default=0, type=int,
        help=('Put an N into sites where the coverage is below the specified '
              'cutoff. If you specify a negative numer, masking will be '
              'turned off. Requires --bam.'))

    parser.add_argument(
        '--log', default=False, action='store_true',
        help=('Show a log of commands that were (or would be, if --dryRun is '
              'used) executed.'))

    parser.add_argument(
        '--noClean', default=True, action='store_false', dest='clean',
        help=('Do not remove intermediate files or the temporary directory.'))

    parser.add_argument(
        '--callHaplotypesGATK', default=False, action='store_true',
        help=('Use GATK to call haplotypes. See '
              'https://gatk.broadinstitute.org for details on GATK.'))

    parser.add_argument(
        '--picardJar',
        help=('The path to the Picard jar file. See '
              'https://github.com/broadinstitute/picard for details on '
              'Picard.'))

    args = parser.parse_args()

    if not (args.bam or args.vcfFile):
        print('At least one of --bam or --vcfFile must be given.',
              file=sys.stderr)
        sys.exit(0)

    if args.maskLowCoverage and not args.bam:
        print('If --maskLowCoverage is used, --bam must be too.',
              file=sys.stderr)
        sys.exit(0)

    e = Executor(args.dryRun)

    tempdir = mkdtemp(prefix='consensus-')

    if args.vcfFile:
        vcfFile = args.vcfFile
    else:
        # No VCF file provided, so make one.
        vcfFile = join(tempdir, 'vcf.gz')
        if args.callHaplotypesGATK:
            e.execute("samtools index '%s'" % args.bam)
            if args.picardJar:
                picardJar = args.picardJar
            else:
                try:
                    picardJar = os.environ['PICARD_JAR']
                except KeyError:
                    print('If you use --callHaplotypesGATK, you must give a '
                          'Picard JAR file with --picardJar or else set '
                          'PICARD_JAR in your environment.', file=sys.stderr)
                    sys.exit(0)

            indexFile = args.reference + '.fai'
            if os.path.exists(indexFile):
                removeIndex = False
            else:
                removeIndex = True
                e.execute("samtools faidx '%s'" % args.reference)

            if args.reference.lower().endswith('.fasta'):
                dictFile = args.reference[:-len('.fasta')] + '.dict'
            else:
                dictFile = args.reference + '.dict'

            if os.path.exists(dictFile):
                removeDict = False
            else:
                removeDict = True
                e.execute(
                    "java -jar '%s' CreateSequenceDictionary R='%s' O='%s'"
                    % (picardJar, args.reference, dictFile))

            e.execute(
                'gatk --java-options -Xmx4g HaplotypeCaller '
                "--reference '%s' "
                "--input '%s' "
                "--output '%s' "
                "--sample-ploidy 1 "
                '-ERC GVCF' %
                (args.reference, args.bam, vcfFile))

            if removeIndex:
                e.execute("rm '%s'" % indexFile)

            if removeDict:
                e.execute("rm '%s'" % dictFile)
        else:
            e.execute("bcftools mpileup --max-depth 5000 -Ou -f '%s' '%s' | "
                      "bcftools call --ploidy 1 -mv -Oz -o '%s'" %
                      (args.reference, args.bam, vcfFile))

            e.execute("bcftools index '%s'" % vcfFile)

    if args.maskLowCoverage >= 0:
        # Make a BED file.
        bedFile = join(tempdir, 'mask.bed')
        # The doubled-% below are so that Python doesn't try to fill in the
        # values and instead just generates a single % that awk sees.
        e.execute(
            "samtools depth -a '%s' | "
            "awk '$3 < %d {printf \"%%s\\t%%d\\t%%d\\n\", "
            "$1, $2 - 1, $2}' > '%s'" %
            (args.bam, args.maskLowCoverage, bedFile))
        maskArg = '--mask ' + bedFile
    else:
        maskArg = ''

    if args.sample:
        sample = args.sample
    else:
        result = e.execute(
            "gunzip -c '%s' | egrep -m 1 '^#CHROM' | cut -f10" % vcfFile)
        sample = result.stdout.strip()

    consensusFile = join(tempdir, 'consensus.fasta')
    result = e.execute(
        "bcftools consensus --sample '%s' --iupac-codes %s --fasta-ref "
        "'%s' '%s' > '%s'" %
        (sample, maskArg, args.reference, vcfFile, consensusFile))

    consensus = list(FastaReads(consensusFile))[0]

    if args.id is not None:
        consensus.id = args.id
    elif args.idLambda is not None:
        idLambda = eval(args.idLambda)
        consensus.id = idLambda(consensus.id)

    print(consensus.toString('fasta'), end='')

    if result.stderr:
        print(result.stderr, end='', file=sys.stderr)

    if args.dryRun or args.log:
        print('\n'.join(e.log), file=sys.stderr)

    if tempdir:
        if args.clean:
            e.execute("rm -r '%s'" % tempdir)
        else:
            print('Temporary directory %r.' % tempdir, file=sys.stderr)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from __future__ import print_function

import sys
from ete3 import Tree
import argparse


def readNames(fp):
    """
    Read the existing / news names file.

    @param fp: A file pointer containing renaming information.

    @raise ValueError: If there is a duplicate old name given or if a line
        cannot be split into two TAB-separated fields.
    @return: A C{dict} of old to new names.
    """
    result = {}
    for line in fp:
        old, new = line.strip().split('\t')
        if old in result:
            raise ValueError('Sequence id %r appears twice in the input.' %
                             old)
        else:
            result[old] = new

    return result


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Relabel a Newick tree according to a file of from/to '
                 'names.'))

parser.add_argument(
    'newickFile',
    help='The Newick tree file.')

parser.add_argument(
    '--nameFile',
    help=('The file of renamings. Each line should contain an existing (in '
          'the tree) sequence id, a TAB, and then a new sequence id. If not '
          'given, the renaming is read from standard input.'))

parser.add_argument(
    '--inputFormat', type=int, default=0,
    help=('The format argument to pass to the ete3 Tree class initializer to '
          'read the input Newick.'))

parser.add_argument(
    '--outputFormat', type=int, default=0,
    help=('The format argument to pass to the ete3 Tree writer for the '
          'resulting tree.'))

parser.add_argument(
    '--quotedNames', default=False, action='store_true',
    help=('If given, pass quoted_node_names=True to the ete3 Tree class '
          'initializer'))

parser.add_argument(
    '--verbose', default=False, action='store_true',
    help='If given, print details of the renaming(s).')

args = parser.parse_args()

tree = Tree(args.newickFile, format=args.inputFormat,
            quoted_node_names=args.quotedNames)

if args.nameFile:
    with open(args.nameFile) as fp:
        names = readNames(fp)
else:
    names = readNames(sys.stdin)


# Relabel tree nodes that are mentioned in args.nameFile.
for node in tree.traverse():
    try:
        newName = names[node.name]
    except KeyError:
        pass
    else:
        if args.verbose:
            print('Renaming %s to %s' % (node.name, newName), file=sys.stderr)
        node.name = newName

# Print the tree, including the name of the root.
print('%s%s;' % (tree.write(format=args.outputFormat)[:-1], tree.name))



#!/usr/bin/env python

from setuptools import setup


# Modified from http://stackoverflow.com/questions/2058802/
# how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
def version():
    import os
    import re

    init = os.path.join('dark', '__init__.py')
    with open(init) as fp:
        initData = fp.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]+)['\"]",
                      initData, re.M)
    if match:
        return match.group(1)
    else:
        raise RuntimeError('Unable to find version string in %r.' % init)


# Explicitly list bin scripts to be installed, seeing as I have a few local
# bin files that are not (yet) part of the distribution.
scripts = [
    'bin/aa-info.py',
    'bin/aa-to-dna.py',
    'bin/aa-to-properties.py',
    'bin/adaptor-distances.py',
    'bin/alignment-panel-civ.py',
    'bin/alignments-per-read.py',
    'bin/bit-score-to-e-value.py',
    'bin/cat-json-blast-records.py',
    'bin/check-fasta-json-blast-consistency.py',
    'bin/codon-distance.py',
    'bin/compare-consensuses.py',
    'bin/compare-sequences.py',
    'bin/convert-blast-xml-to-json.py',
    'bin/convert-diamond-to-json.py',
    'bin/convert-diamond-to-sam.py',
    'bin/convert-sam-to-fastq.sh',
    'bin/create-newick-relabeling-output.py',
    'bin/dark-matter-version.py',
    'bin/describe-protein-database.py',
    'bin/dna-to-aa.py',
    'bin/download-genbank.sh',
    'bin/e-value-to-bit-score.py',
    'bin/extract-ORFs.py',
    'bin/fasta-base-indices.py',
    'bin/fasta-count.py',
    'bin/fasta-diff.sh',
    'bin/fasta-identity-table.py',
    'bin/fasta-ids.py',
    'bin/fasta-join.py',
    'bin/fasta-lengths.py',
    'bin/fasta-sequences.py',
    'bin/fasta-sort.py',
    'bin/fasta-split-by-id.py',
    'bin/fasta-subset.py',
    'bin/fasta-subtraction.py',
    'bin/fasta-to-phylip.py',
    'bin/fasta-variable-sites.py',
    'bin/filter-fasta-by-complexity.py',
    'bin/filter-fasta-by-taxonomy.py',
    'bin/filter-fasta.py',
    'bin/filter-hits-to-fasta.py',
    'bin/filter-reads-alignments.py',
    'bin/filter-sam.py',
    'bin/find-hits.py',
    'bin/format-fasta.py',
    'bin/genome-protein-summary.py',
    'bin/get-features.py',
    'bin/get-hosts.py',
    'bin/get-reads.py',
    'bin/get-taxonomy.py',
    'bin/graph-evalues.py',
    'bin/local-align.py',
    'bin/make-consensus.py',
    'bin/make-fasta-database.py',
    'bin/make-protein-database.py',
    'bin/ncbi-fetch-id.py',
    'bin/newick-to-ascii.py',
    'bin/noninteractive-alignment-panel.py',
    'bin/parse-genbank-flat-file.py',
    'bin/position-summary.py',
    'bin/pre-commit.sh',
    'bin/print-blast-xml-for-derek.py',
    'bin/print-blast-xml.py',
    'bin/print-read-lengths.py',
    'bin/proteins-to-pathogens.py',
    'bin/proteins-to-pathogens-civ.py',
    'bin/randomize-fasta.py',
    'bin/read-blast-json.py',
    'bin/read-blast-xml.py',
    'bin/relabel-newick-tree.py',
    'bin/run-bwa.py',
    'bin/run-bowtie2.py',
    'bin/sam-coverage.py',
    'bin/sam-coverage-depth.py',
    'bin/sam-to-fasta-alignment.py',
    'bin/sam-reference-read-counts.py',
    'bin/sam-references.py',
    'bin/sff-to-fastq.py',
    'bin/split-fasta-by-adaptors.py',
    'bin/subset-protein-database.py',
    'bin/summarize-fasta-bases.py',
    'bin/summarize-reads.py',
    'bin/trim-primers.py',
    'bin/trim-reads.py',
    'bin/write-htcondor-job-spec.py',
]

setup(name='dark-matter',
      version=version(),
      packages=['dark', 'dark.blast', 'dark.diamond', 'dark.civ'],
      url='https://github.com/acorg/dark-matter',
      download_url='https://github.com/acorg/dark-matter',
      author='Terry Jones, Barbara Muehlemann, Tali Veith, Sophie Mathias',
      author_email='tcj25@cam.ac.uk',
      keywords=['virus discovery'],
      classifiers=[
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      license='MIT',
      description='Python classes for working with genetic sequence data',
      scripts=scripts,
      install_requires=[
          'biopython>=1.71',
          'bz2file>=0.98',
          'Cython>=0.29.16',
          'ipython>=3.1.0',
          'matplotlib>=1.4.3',
          'mysql-connector-python==8.0.11',
          'numpy>=1.14.2',
          'pysam>=0.15.2',
          'pyfaidx>=0.4.8.4',
          'pyzmq>=14.3.1',
          'requests>=2.18.4',
          'cachetools>=3.1.0',
          'simplejson>=3.5.3',
          'six>=1.11.0',
      ])



from unittest import TestCase
from six import assertRaisesRegex

from dark.html import NCBISequenceLinkURL, NCBISequenceLink


class TestNCBISequenceLinkURL(TestCase):
    """
    Test the NCBISequenceLinkURL function.
    """

    def testNoField(self):
        """
        If no field is passed, the passed title must be used.
        """
        self.assertEqual('http://www.ncbi.nlm.nih.gov/nuccore/xxx',
                         NCBISequenceLinkURL('xxx'))

    def testFieldNumber(self):
        title = 'gi|323924|gb|M15204.1|FCVMYCCA Feline leukemia virus myc gene'
        self.assertEqual('http://www.ncbi.nlm.nih.gov/nuccore/M15204.1',
                         NCBISequenceLinkURL(title, 3))

    def testAlternateDelimiter(self):
        title = 'gi+37955203+gb+AY253278.1+ Homo sapiens clone AL-11 HIV-1'
        self.assertEqual('http://www.ncbi.nlm.nih.gov/nuccore/AY253278.1',
                         NCBISequenceLinkURL(title, 3, '+'))

    def testFieldOutOfRange(self):
        """
        If the field number passed does not exist, IndexError must be raised.
        """
        title = 'gi|37955203|gb|AY253278.1| Homo sapiens clone AL-11 HIV-1'
        error = (
            r"^Could not extract field 10 from sequence title "
            r"'gi\|37955203\|gb\|AY253278\.1\| Homo sapiens clone "
            "AL-11 HIV-1'$")
        assertRaisesRegex(self, IndexError, error, NCBISequenceLinkURL, title,
                          10)


class TestNCBISequenceLink(TestCase):
    """
    Test the NCBISequenceLink function.
    """
    def testNoField(self):
        """
        If no field is passed, the passed title must be used.
        """
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/xxx" ' +
            'target="_blank">xxx</a>', NCBISequenceLink('xxx'))

    def testFieldNumber(self):
        title = 'gi|323924|gb|M15204.1|FCVMYCCA Feline leukemia virus myc gene'
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/M15204.1" ' +
            'target="_blank">' + title + '</a>', NCBISequenceLink(title, 3))

    def testAlternateDelimiter(self):
        title = 'gi+37955203+gb+AY253278.1+ Homo sapiens clone AL-11 HIV-1'
        self.assertEqual(
            '<a href="http://www.ncbi.nlm.nih.gov/nuccore/AY253278.1" ' +
            'target="_blank">' + title + '</a>',
            NCBISequenceLink(title, 3, '+'))

    def testFieldOutOfRange(self):
        """
        If the field number passed does not exist, IndexError must be raised.
        """
        title = 'gi|37955203|gb|AY253278.1| Homo sapiens clone AL-11 HIV-1'
        error = (
            r"^Could not extract field 10 from sequence title "
            r"'gi\|37955203\|gb\|AY253278\.1\| Homo sapiens clone "
            "AL-11 HIV-1'$")
        assertRaisesRegex(self, IndexError, error, NCBISequenceLink, title, 10)



#!/usr/bin/env python

from __future__ import print_function

import sys

from dark.aa import find
from dark.aa import CODONS

from dark.aa import AA_LETTERS, ALL_PROPERTIES, PROPERTY_NAMES

args = AA_LETTERS if len(sys.argv) == 1 else sys.argv[1:]
error = False

for arg in args:
    aas = list(find(arg))
    if aas:
        for aa in aas:
            print(aa.name)
            print('  3-letter abbreviation: %s' % aa.abbrev3)
            print('  1-letter abbreviation: %s' % aa.abbrev1)
            print('  Codons: %s' % ', '.join(sorted(aa.codons)))

            properties = []
            print('  Properties:', end=' ')
            for prop in ALL_PROPERTIES:
                if aa.properties & prop:
                    properties.append(PROPERTY_NAMES[prop])
            print(', '.join(properties))

            print('  Property details:')
            for propertyDetail, value in aa.propertyDetails.items():
                print('    %s: %s' % (propertyDetail, value))
    else:
        error = True
        print('Unknown amino acid or codon: %s' % arg, file=sys.stderr)

if error:
    print('Valid arguments are: %s.' % list(CODONS), file=sys.stderr)



from __future__ import division, print_function

import six
from time import time, ctime

from subprocess import PIPE, CalledProcessError
if six.PY3:
    from subprocess import run
else:
    from subprocess import check_call


class Executor(object):
    """
    Log and execute shell commands.

    @param dryRun: If C{True}, do not execute commands, just log them.
        This sets the default and can be overidden for a specific command
        by passing C{dryRun} to the C{execute} method.
    """
    def __init__(self, dryRun=False):
        self.dryRun = dryRun
        self.log = [
            '# Executor created at %s. Dry run = %s.' % (ctime(time()), dryRun)
        ]

    def dryRun(self):
        """
        Is this a dry run?

        @return: A Boolean indicating whether this is a dry run.
        """
        return self._dryRun

    def execute(self, command, dryRun=None, useStderr=True, **kwargs):
        """
        Execute (or simulate) a command. Add to our log.

        @param command: Either a C{str} command (which will be passed to the
            shell) or a C{list} of command arguments (including the executable
            name), in which case the shell is not used.
        @param dryRun: If C{True}, do not execute commands, just log them.
            If C{False}, execute the commands. If not given or C{None}, use
            the default setting (in C{self.dryRun}).
        @param useStderr: If C{True} print a summary of the command standard
            output and standard error to sys.stderr if the command results in
            an exception. If a function is passed, the exception is passed to
            the function and the summary is printed to sys.stderr if the
            function returns C{True}.
        @param kwargs: Keyword arguments that will be passed to subprocess.run
            (or subprocess.check_call for Python version 2). Note that keyword
            arguments are not currently logged (the logging is slightly
            problematic since a keyword argument might be an environment
            dictionary).
        @raise CalledProcessError: If the command results in an error.
        @return: A C{CompletedProcess} instance. This has attributes such as
            C{returncode}, C{stdout}, and C{stderr}. See pydoc subprocess.
            If C{dryRun} is C{True}, C{None} is returned.
        """
        if isinstance(command, six.string_types):
            # Can't have newlines in a command given to the shell.
            strCommand = command = command.replace('\n', ' ').strip()
            shell = True
        else:
            strCommand = ' '.join(command)
            shell = False

        dryRun = self.dryRun if dryRun is None else dryRun

        if dryRun:
            self.log.append('$ ' + strCommand)
            return

        start = time()
        self.log.extend([
            '# Start command (shell=%s) at %s' % (shell, ctime(start)),
            '$ ' + strCommand,
        ])

        if six.PY3:
            try:
                result = run(command, check=True, stdout=PIPE, stderr=PIPE,
                             shell=shell, universal_newlines=True, **kwargs)
            except CalledProcessError as e:
                if callable(useStderr):
                    useStderr = useStderr(e)
                if useStderr:
                    import sys
                    print('CalledProcessError:', e, file=sys.stderr)
                    print('STDOUT:\n%s' % e.stdout, file=sys.stderr)
                    print('STDERR:\n%s' % e.stderr, file=sys.stderr)
                raise
        else:
            try:
                result = check_call(command, stdout=PIPE, stderr=PIPE,
                                    shell=shell, universal_newlines=True,
                                    **kwargs)
            except CalledProcessError as e:
                if callable(useStderr):
                    useStderr = useStderr(e)
                if useStderr:
                    import sys
                    print('CalledProcessError:', e, file=sys.stderr)
                    print('Return code: %s' % e.returncode, file=sys.stderr)
                    print('Output:\n%s' % e.output, file=sys.stderr)
                raise

        stop = time()
        elapsed = (stop - start)
        self.log.extend([
            '# Stop command at %s' % ctime(stop),
            '# Elapsed = %f seconds' % elapsed,
        ])

        return result



#!/usr/bin/env python

"""
Read DNA or AA FASTA from stdin and print AA ORF-only FASTA to stdout. I.e.,
the output FASTA will contain one sequence for each ORF found in each
sequence on stdin.

If a minimal ORF length is given, only print ORFs of at least that length
unless we are asked to allow ORFs that are open at at least one end.

Note that no start or stop codons will appear in the output. If you are
wanting to simply translate DNA FASTA to AA FASTA on a sequence-by-sequence
basis leaving in start/stop codons, use dna-to-aa.py instead.
"""

from __future__ import print_function

import sys
from os.path import basename
import argparse

from Bio.Data.CodonTable import TranslationError

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Convert DNA or AA FASTA to AA ORF-only FASTA',
        epilog=('Given DNA or AA FASTA on stdin, output AA ORF-only FASTA to '
                'stdout. Optionally, filter by minimum required ORF length '
                'and allow the output of short ORFs that are open.'))

    parser.add_argument(
        '--allowOpenORFs', default=False, action='store_true',
        help=('If True, ORFs that do not meet the length requirement '
              '(as specified by --minORFLength) will be output as long as '
              'they are open.'))

    parser.add_argument(
        '--minORFLength', metavar='LEN', type=int, default=None,
        help='Only ORFs of at least this length will be written to stdout.')

    parser.add_argument(
        '--maxORFLength', metavar='LEN', type=int, default=None,
        help=('Only ORFs of a maximum of this length will be written to '
              'stdout.'))

    parser.add_argument(
        '--kozakOnly', default=False, action='store_true',
        help=('Only ORFs that also have a Kozak consensus will be written to '
              'stdout. Only applicable if DNA reads are given.'))

    parser.add_argument(
        '--kozakInfoFile',
        help=('Filename to which all Kozak consensus information is written. '
              'Only applicable if DNA reads are given.'))

    addFASTACommandLineOptions(parser)

    args = parser.parse_args()
    allowOpenORFs = args.allowOpenORFs
    minORFLength = args.minORFLength
    maxORFLength = args.maxORFLength
    kozakInfoFile = args.kozakInfoFile
    kozakOnly = args.kozakOnly
    reads = parseFASTACommandLineOptions(args)
    aa = args.readClass in ('AARead', 'AAReadORF', 'AAReadWithX', 'SSAARead',
                            'SSAAReadWithX', 'TranslatedRead')

    if aa:
        def translations(read):
            return (read,)
    else:
        def translations(read):
            return read.translations()

    if kozakInfoFile or kozakOnly:
        if aa:
            print('Kozak sequences cannot be computed from aa sequences.',
                  file=sys.stderr)
            sys.exit(1)
        else:
            from dark.dna import findKozakConsensus

        def writeToKozakOut(kozakread, kozakfp):
            """
            Writes out information about a Kozak sequence stored in kozakread
            to the file name given in kozakfp.
            """
            print('Read ID: ' + kozakread.id, file=kozakfp)
            print('Kozak sequence: ' + kozakread.sequence, file=kozakfp)
            print('Offset of the Kozak sequence A: ' +
                  str(kozakread.start + 6), file=kozakfp)
            print('Quality of the Kozak consensus: ' +
                  str(kozakread.kozakQuality) + ' %', file=kozakfp)

    def findORFs(kozakfp):
        for read in reads:
            try:
                for translation in translations(read):
                    for orf in translation.ORFs(allowOpenORFs):
                        # Check the length requirements, if any.
                        if ((minORFLength is None or
                             len(orf) >= minORFLength) and
                            (maxORFLength is None or
                             len(orf) <= maxORFLength)):

                            if kozakOnly:
                                for kozakread in findKozakConsensus(read):
                                    start = orf.start * 3
                                    if (start + 1 <=
                                       kozakread.stop <= start + 3):
                                        print(orf.toString('fasta'),
                                              end='')
                            else:
                                print(orf.toString('fasta'), end='')
                            if kozakfp:
                                for kozakread in findKozakConsensus(read):
                                    start = orf.start * 3
                                    if (start + 1 <=
                                       kozakread.stop <= start + 3):
                                        writeToKozakOut(kozakread, kozakfp)

            except TranslationError as error:
                print('Could not translate read %r sequence %r (%s).' %
                      (read.id, read.sequence, error), file=sys.stderr)
                if not aa:
                    print('Did you forget to run '
                          '%s with "--readClass AARead"?' %
                          (basename(sys.argv[0])), file=sys.stderr)
                sys.exit(1)

    if kozakInfoFile:
        with open(kozakInfoFile, 'a+') as kozakfp:
            findORFs(kozakfp)
    else:
        findORFs(None)
